"""Matchers for comparing to strings."""

# pylint: disable=too-few-public-methods

import re

from h_matchers.matcher.core import Matcher

__all__ = ["AnyString", "AnyStringContaining", "AnyStringMatching"]


class AnyStringContaining(Matcher):
    """Matches any string with a certain substring."""

    def __init__(self, sub_string):
        super().__init__(
            f"*{sub_string}*",
            lambda other: isinstance(other, str) and sub_string in other,
        )


class AnyStringMatching(Matcher):
    """Matches any regular expression."""

    def __init__(self, pattern, flags=0):
        """Create a string matcher with the specified regex.

        :param pattern: The raw pattern to compile into a regular expression
        :param flags: Flags `re` e.g. `re.IGNORECASE`
        """
        regex = re.compile(pattern, flags)
        super().__init__(
            pattern, lambda other: isinstance(other, str) and regex.match(other)
        )


class AnyString(Matcher):
    """Matches any string."""

    matching = AnyStringMatching
    containing = AnyStringContaining

    def __init__(self):
        super().__init__("* any string *", lambda other: isinstance(other, str))



"""A mixin for AnyCollection which lets you check for specific items."""

from h_matchers.decorator import fluent_entrypoint
from h_matchers.exception import NoMatch
from h_matchers.matcher.collection.containment import (
    AnyIterableWithItems,
    AnyIterableWithItemsInOrder,
    AnyMappingWithItems,
)


class ContainsMixin:
    """Check specific items are in the container."""

    _items = None
    _in_order = False
    _exact_match = False

    @staticmethod
    def containing(items):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @fluent_entrypoint
    # pylint: disable=function-redefined
    def containing(self, items):
        """Specify that this item must contain these items.

        By default we will attempt to match the items in any order.
        If you want to change this you can call `in_order()`.

        If a list of items is provided then these will be checked in order, if
        the comparison object supports ordering.

        If a dict of items is provided, then both the keys and values will be
        checked to see if they match.

        Can be called as an instance or class method.

        :param items: A set or list of items to check for
        :raises ValueError: If you provide something other than a set or list
        """

        if isinstance(items, ContainsMixin):
            # It's ok, because it's our class
            # pylint: disable=protected-access
            items = items._items

        self._items = items

    def in_order(self):
        """Set that matched items can occur in any order.

        :raises ValueError: If no items have been set
        :return: self - for fluent chaining
        """
        if self._items is None:
            raise ValueError("You must set items before calling this")

        self._in_order = True
        return self

    def only(self):
        """Set that only the provided items should be in the collection.

        :raises ValueError: If not items have been set
        :return: self - for fluent chaining
        """
        if self._items is None:
            raise ValueError("You must set items before calling this")

        self._exact_match = True

        return self

    def _check_contains(self, other, original=None):
        if not self._items:
            return

        # We can bail out early if we need an exact match and they are
        # different sizes
        if self._exact_match and len(self._items) != len(other):
            raise NoMatch("Items of different size")

        if hasattr(self._items, "items"):
            matcher_class = AnyMappingWithItems
        elif self._in_order:
            matcher_class = AnyIterableWithItemsInOrder
        else:
            matcher_class = AnyIterableWithItems

        if matcher_class(self._items) != original:
            raise NoMatch()

    def _describe_contains(self):
        if not self._items:
            return

        yield "containing"
        if self._exact_match:
            yield "only"

        yield f"{self._items}"

        if self._in_order:
            yield "in order"



"""A fluent interface over AnyURLCore for matching URLs."""
from h_matchers.decorator import fluent_entrypoint
from h_matchers.matcher.collection import AnyMapping
from h_matchers.matcher.strings import AnyString
from h_matchers.matcher.web.url.core import AnyURLCore


class AnyURL(AnyURLCore):
    """A URL matcher with a fluent style interface."""

    # pylint: disable=function-redefined

    PRESENT_DEFAULT = {
        "scheme": AnyString(),
        "host": AnyString(),
        "path": AnyString(),
        "query": AnyMapping(),
        "fragment": AnyString(),
    }

    def _apply_field_default(self, field, value):
        if value is not AnyURLCore.APPLY_DEFAULT:
            return value

        return self.PRESENT_DEFAULT[field]

    @staticmethod
    def with_scheme(scheme=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @staticmethod
    def with_host(host=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @staticmethod
    def with_path(path=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @staticmethod
    def containing_query(query=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @staticmethod
    def with_query(query=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @staticmethod
    def with_fragment(fragment=AnyURLCore.APPLY_DEFAULT):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @classmethod
    def matching(cls, base_url):
        """Create a URL matcher based on the given URL.

        :param base_url: URL to base the matcher on
        :return: An instance of AnyURLFluent
        """
        return AnyURL(base_url)

    @fluent_entrypoint
    def with_scheme(self, scheme=AnyURLCore.APPLY_DEFAULT):
        """Specify that this URL must have a scheme or None.

        If you pass None this will ensure the URL has no scheme.

        :param scheme: None, string or matcher for the scheme
        """

        self.parts["scheme"] = self._apply_field_default("scheme", scheme)

    @fluent_entrypoint
    def with_host(self, host=AnyURLCore.APPLY_DEFAULT):
        """Specify that this URL must have a host or None.

        If you pass None this will ensure the URL has no host.

        :param host: None, string or matcher for the host
        """
        self.parts["host"] = self._apply_field_default("host", host)

    @fluent_entrypoint
    def with_path(self, path=AnyURLCore.APPLY_DEFAULT):
        """Specify that this URL must have a path or None.

        If you pass None this will ensure the URL has no path.

        :param path: None, string or matcher for the path
        """
        if path is AnyURLCore.APPLY_DEFAULT:
            self.parts["path"] = AnyString()
        else:
            self.parts["path"] = self._get_path_matcher(
                path, self.parts["scheme"], self.parts["host"]
            )

    @fluent_entrypoint
    def containing_query(self, query):
        """Specify that the query must have at least the items specified.

        :param query: A mappable to check
        """
        self._set_query(query, exact_match=False)

    @fluent_entrypoint
    def with_query(self, query=AnyURLCore.APPLY_DEFAULT):
        """Specify that this URL must have a query or None.

        If you pass None this will ensure the URL has no query.

        :param query: None, mapping or matcher for the query
        """
        query = self._apply_field_default("query", query)
        self._set_query(query)

    @fluent_entrypoint
    def with_fragment(self, fragment=AnyURLCore.APPLY_DEFAULT):
        """Specify that this URL must have a fragment or None.

        If you pass None this will ensure the URL has no fragment.

        :param fragment: None, string or matcher for the fragment
        """
        self.parts["fragment"] = self._apply_field_default("fragment", fragment)



"""The core classes for matching.

These are not intended to be used directly.
"""

# pylint: disable=too-few-public-methods


class Matcher:
    """Used as the base class for concrete matching classes.

    Implements a base class for use in the testing pattern where an object
    stands in for another and will evaluate to true when compared with the
    other.
    """

    def __init__(self, description, test_function):
        self._description = description
        self._test_function = test_function

    def __eq__(self, other):
        return self._test_function(other)

    def __str__(self):
        return self._description  # pragma: no cover

    def __repr__(self):
        return f"<{self.__class__.__name__} '{str(self)}'>"  # pragma: no cover



"""A collection of matchers for various number types."""

# pylint: disable=too-few-public-methods

from h_matchers.matcher.core import Matcher


class AnyInt(Matcher):
    """Matches any integer."""

    def __init__(self):
        super().__init__(
            "* any int *",
            lambda other: isinstance(other, int) and other is not True
            # pylint: disable=compare-to-zero
            and other is not False,
        )



"""A mixin for AnyCollection which lets you specify the type."""

# pylint: disable=too-few-public-methods

from h_matchers.decorator import fluent_entrypoint
from h_matchers.exception import NoMatch


class TypeMixin:
    """Apply and check type constraint."""

    _exact_type = None

    @staticmethod
    def of_type(of_type):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @fluent_entrypoint
    # pylint: disable=function-redefined
    def of_type(self, of_type):
        """Limit the type to a specific type like `list` or `set`.

        Can be called as an instance or class method.
        """
        self._exact_type = of_type

    def _check_type(self, _, original):
        if self._exact_type:
            if not isinstance(original, self._exact_type):
                raise NoMatch("Wrong type")

    def _describe_type(self):
        if self._exact_type:
            yield self._exact_type.__name__
        else:
            yield "iterable"



"""The public interface class for comparing with things."""
from h_matchers.matcher.anything import AnyThing
from h_matchers.matcher.collection import (
    AnyCollection,
    AnyDict,
    AnyList,
    AnyMapping,
    AnySet,
)
from h_matchers.matcher.combination import AllOf, AnyOf
from h_matchers.matcher.meta import AnyCallable, AnyFunction
from h_matchers.matcher.number import AnyInt
from h_matchers.matcher.object import AnyObject
from h_matchers.matcher.strings import AnyString
from h_matchers.matcher.web.request import AnyRequest
from h_matchers.matcher.web.url import AnyURL

# pylint: disable=too-few-public-methods

__all__ = ["Any", "All"]


class Any(AnyThing):
    """Matches anything and provides access to other matchers."""

    # pylint: disable=too-few-public-methods

    string = AnyString
    int = AnyInt
    object = AnyObject

    function = AnyFunction
    callable = AnyCallable

    iterable = AnyCollection
    mapping = AnyMapping
    list = AnyList
    set = AnySet
    dict = AnyDict

    url = AnyURL
    request = AnyRequest

    of = AnyOf

    @staticmethod
    def instance_of(type_):
        """Specify that this item must be an instance of the provided type.

        :return: An instance of AnyObject configured with the given type.
        """
        return AnyObject.of_type(type_)


class All(AllOf):
    """Matches when all items match.

    Mostly a sop to create a consistent interface.
    """

    of = AllOf



"""Matchers for simple objects."""
from h_matchers.decorator import fluent_entrypoint
from h_matchers.matcher.core import Matcher

# pylint: disable=function-redefined


class AnyObject(Matcher):  # pragma: no cover
    """Match any object, optionally with a specific class or attributes.

    It's possible to instantiate this object without specifying a type or any
    attributes, but as most everything is an object in Python 3, this is
    largely equivalent to an `Any()` matcher.

    Any specified attributes must be present on the object being compared to,
    and must have matching values. It's is possible to nest matchers if you
    require an attribute to be present, but not have any particular value:

        AnyObject.with_attrs({"name": Any()})

    Any attributes specified will be available as attributes on the matcher
    object. This is helpful if you need to sort the items before comparing them
    for example.
    """

    def __init__(self, type_=None, attributes=None):
        """Create a new object matcher.

        :param type_: The type this object will match
        :param attributes:  A mapping of attributes to values to match
        """
        # Use scrambled names to reduce the chances of attribute clashes
        self.__type = type_
        self.__attributes = attributes

        super().__init__("dummy", self._matches_object)

    @staticmethod
    def of_type(type_):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @fluent_entrypoint
    def of_type(self, type_):
        """Specify that this item must be an instance of the provided type.

        Can be called as an instance or class method.

        :param type_: The type this object will match
        """
        self.__type = type_

    @staticmethod
    def with_attrs(attributes):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @fluent_entrypoint
    def with_attrs(self, attributes):
        """Specify that this item must have at least these attribute values.

        Takes a mapping containing the names of attributes that the target
        object must have, and values that those values must take.

        Any attributes specified this way are accessible on this matcher as
        attributes, unless they clash with methods of this class.

        Can be called as an instance or class method.

        :param attributes: A mapping of attributes to values
        :raise ValueError: If the provided attributes do not support `items()`
        """
        if not hasattr(attributes, "items"):
            raise ValueError("The attributes must be a mapping")

        self.__attributes = attributes

    def _matches_object(self, other):
        if self.__type is not None and not isinstance(other, self.__type):
            return False

        if self.__attributes is not None:
            for key, value in self.__attributes.items():
                if not hasattr(other, key):
                    return False

                if getattr(other, key) != value:
                    return False

        return True

    def __getattr__(self, item):
        """Allow our attributes spec to be accessed as attributes."""

        if self.__attributes is not None and item in self.__attributes:
            return self.__attributes[item]

        return super().__getattribute__(item)

    def __str__(self):
        extras = f" with attributes {self.__attributes}" if self.__attributes else ""

        instance = object.__name__ if self.__type is None else self.__type.__name__

        return f"<Any instance of '{instance}'{extras}>"



"""A matcher for URLs.

## Basic initialisation

If no arguments are given, a matcher that matches any valid URL will be
initialized:

    AnyURL()

There's a good chance you _don't_ want this however, as actual URLs are much
more flexible than you might assume. Basically any string will match.

You can limit what the URL will match by specifying the components you want to
limit:

    AnyURL(scheme='https', query={'a': 2})

## Based on a URL

If `base_url` is given it will be split into its component parts (scheme,
host, path, etc.) and a matcher based on those components being an exact match
will be initialized:

    AnyURL("http://example.com/path?a=b#anchor")

You can override any particular part in combination with a base URL:

    AnyURL("http://example.com/path", scheme=Any.of(['http', 'https']))

## Specifying a URL query

There are many ways you can specify a URL query to match:

    AnyURL(query='a=1&b=2')
    AnyURL(query={'a': '1', 'b': '2'})
    AnyURL(query=Any.mapping.containing({'a': '2'}))

Any mappable object can be used as the specification including mutli-valued
dicts like ``webob.multidict.MultiDict``. A string with multiple values for
the same key can also be used.

# Path matching

Comparing paths is tricky, particularly when comparing to bare paths without
a scheme or host. The following rules are in place to try and make this easier:

 * If you specify `host=None`, `scheme=None` the path is matched exactly as you give it
 * In all other cases the path will match with or without a leading slash

For example:

    Any.url(scheme=None, host=None, path="foo") == "/foo"  # False
    Any.url(scheme=None, host=None, path="/foo") == "foo"  # False

    Any.url(path="/foo") == "http://example.com/foo"       # True
    Any.url(path="foo") == "http://example.com/foo"        # True
    Any.url(path="/foo") == "foo"                          # True
    Any.url(path="foo") == "/foo"                          # True
"""
import re
from collections import Counter
from urllib.parse import parse_qsl, urlparse

from h_matchers.matcher.collection import AnyMapping
from h_matchers.matcher.combination import AnyOf, NamedMatcher
from h_matchers.matcher.core import Matcher
from h_matchers.matcher.strings import AnyString, AnyStringMatching

# pylint: disable=too-few-public-methods,no-value-for-parameter


class AnyURLCore(Matcher):
    """Matches any URL."""

    APPLY_DEFAULT = object()
    STRING_OR_NONE = NamedMatcher("<AnyStringOrNone>", AnyOf([None, AnyString()]))
    MAP_OR_NONE = NamedMatcher("<AnyMappingOrNone>", AnyOf([None, AnyMapping()]))

    DEFAULTS = {
        "scheme": STRING_OR_NONE,
        "host": STRING_OR_NONE,
        "path": STRING_OR_NONE,
        "query": MAP_OR_NONE,
        "fragment": STRING_OR_NONE,
    }

    # pylint: disable=too-many-arguments
    # I can't see a way around it. We could use kwargs, but then auto complete
    # would be hard
    def __init__(
        self,
        base_url=None,
        scheme=APPLY_DEFAULT,
        host=APPLY_DEFAULT,
        path=APPLY_DEFAULT,
        query=APPLY_DEFAULT,
        fragment=APPLY_DEFAULT,
    ):
        """Initialize a new URL matcher.

        :param base_url: URL (with scheme) to base the matcher on
        :param scheme: Scheme to match (e.g. http)
        :param host: Hostname to match
        :param path: URL path to match
        :param query: Query to match (string, dict or matcher)
        :param fragment: Anchor fragment to match (e.g. "name" for "#name")
        """
        self.parts = {
            # https://tools.ietf.org/html/rfc7230#section-2.7.3
            # scheme and host are case-insensitive
            "scheme": self._lower_if_string(scheme),
            "host": self._lower_if_string(host),
            # `path`, `query` and `fragment` are case-sensitive
            "path": self._get_path_matcher(path, scheme, host),
            "fragment": fragment,
        }

        self._set_query(query)

        if base_url:
            self._set_base_url(base_url)
        else:
            # Apply default matchers for everything not provided
            self._apply_defaults(self.parts, self.DEFAULTS)

        super().__init__("dummy", self._matches_url)

    def __str__(self):
        contraints = {
            key: value
            for key, value in self.parts.items()
            if value is not self.DEFAULTS[key]
        }

        if not contraints:
            return "* any URL *"

        return f"* any URL matching {contraints} *"

    @classmethod
    def parse_url(cls, url_string):
        """Parse a URL into a dict for comparison.

        Parses the given URL allowing you to see how AnyURL will understand it.
        This can be useful when debugging why a particular URL does or does
        not match.

        :param url_string: URL to parse
        :raise ValueError: If scheme is mandatory and not provided
        :return: A normalised string of comparison values
        """
        url = urlparse(url_string)

        if not url.scheme and not url.netloc:
            # Without a scheme `urlparse()` assumes that the hostname is part
            # of the path, so we have to try and guess what it really was

            host, path = cls._guess_hostname_and_path(url.path)
            url = url._replace(netloc=host, path=path)

        return {
            "scheme": url.scheme.lower() if url.scheme else None,
            "host": url.netloc.lower() if url.netloc else None,
            "path": url.path or None,
            "query": MultiValueQuery.normalise(url.query),
            "fragment": url.fragment or None,
        }

    @staticmethod
    def _get_path_matcher(path, scheme, host):
        # If we are anything other than a plain string or None, use it directly
        if path is not None and not isinstance(path, str):
            return path

        # If we are matching paths alone, just return whatever we were given
        # so we match exactly. This lets the user distinguish between /path
        # and path which may be important
        if scheme is None and host is None:
            return path

        # If we got None, we need to allow ourselves to match either slash, ''
        # or None
        if path in (None, "/", ""):
            return NamedMatcher("<Path '/'>", AnyOf([None, AnyStringMatching(r"^/?$")]))

        # Otherwise construct a matcher which doesn't care about leading
        # slashes
        return NamedMatcher(
            f"'<Path '{path}'>", AnyStringMatching(f"^/?{re.escape(path)}$")
        )

    def _set_query(self, query, exact_match=True):
        if query is not self.APPLY_DEFAULT:
            query = MultiValueQuery.normalise(query)
            if query and not isinstance(query, Matcher):
                # MultiValueQuery is guaranteed to return something we can
                # provide to AnyMapping for comparison
                query = AnyMapping.containing(query)
                if exact_match:
                    query = query.only()

        self.parts["query"] = query

    def _set_base_url(self, base_url):
        # If we have a base URL, we'll take everything from there if it
        # wasn't explicitly provided in the constructor

        overlay = self.parse_url(base_url)

        path_matcher = self._get_path_matcher(
            overlay["path"], overlay["scheme"], overlay["host"]
        )
        overlay["path"] = path_matcher
        self._apply_defaults(self.parts, overlay)

    @staticmethod
    def _lower_if_string(value):
        if isinstance(value, str):
            return value.lower()

        return value

    @staticmethod
    def _apply_defaults(values, defaults, default_key=APPLY_DEFAULT):
        for key, default_value in defaults.items():
            if values[key] is default_key:
                values[key] = default_value

    @classmethod
    def _is_hostname(cls, host):
        if not host:
            return False

        if "." in host:
            return True

        return host.lower() == "localhost"

    @classmethod
    def _guess_hostname_and_path(cls, path):
        if "/" in path:
            head, tail = path.split("/", 1)
            if cls._is_hostname(head):
                return head, f"/{tail}"

        elif cls._is_hostname(path):
            return path, None

        return None, path

    def _matches_url(self, other):
        if not isinstance(other, str):
            return False

        comparison = self.parse_url(other)

        return self.parts == comparison


class MultiValueQuery(list):
    """Normalise and represent URL queries.

    This class is for internal use only and should not be used by outside
    consumers.
    """

    def items(self):
        """Iterate over contained items as if a dict.

        The bare minimum to appear as a mapping so that we can be passed to
        `AnyMapping.contains()`.
        """
        yield from self

    @classmethod
    def normalise(cls, query_comparator):
        """Get a normalised form of the representation of a query.

        :return: None, a matcher or something suitable for AnyMapping.
        """
        if query_comparator is None:
            return None

        if isinstance(query_comparator, str):
            return cls._from_query_string(query_comparator)

        return query_comparator

    @classmethod
    def _from_query_string(cls, query_string):
        if not query_string:
            return None

        key_value = parse_qsl(query_string)

        if cls._max_key_repetitions(key_value) > 1:
            return MultiValueQuery(key_value)

        return dict(key_value)

    @classmethod
    def _max_key_repetitions(cls, key_value):
        return Counter(key for key, _ in key_value).most_common(1)[0][1]

    def __repr__(self):
        return f"<MultiValueQuery {super().__repr__()}>"



"""Decorators for h-matchers."""


# Pylint doesn't understand this is a decorator
class fluent_entrypoint:  # pylint: disable=invalid-name,too-few-public-methods
    """Makes a class method act as both a method and a classmethod.

    If the wrapped method is called as a classmethod an instance will first
    be created and then passed to the object. It is therefore important
    that you class not accept any arguments for instantiation.

    This will automatically return `self` for fluent chaining in your methods.
    """

    def __init__(self, function):
        self.function = function

    def __get__(self, obj, _type=None):
        # If we have been called in a classmethod context, create a new
        # instance of the owning object
        if obj is None:
            obj = _type()

        def wrapper(*args, **kwargs):
            self.function(obj, *args, **kwargs)

            return obj

        return wrapper



"""Flexible matchers for various collection types in a fluent style."""

from h_matchers.exception import NoMatch
from h_matchers.matcher.collection import _mixin
from h_matchers.matcher.core import Matcher


class AnyCollection(
    _mixin.SizeMixin,
    _mixin.TypeMixin,
    _mixin.ItemMatcherMixin,
    _mixin.ContainsMixin,
    Matcher,
):
    """Matches any iterable with options for constraining contents and size."""

    def __init__(self):
        # Pass None as our function, as we will be in charge of our own type
        # checking
        super().__init__("dummy", None)

    def __eq__(self, other):
        try:
            copy = list(other)
        except TypeError:
            # Not iterable
            return False

        # Execute checks roughly in complexity order
        for checker in [
            self._check_type,
            self._check_size,
            self._check_item_matcher,
            self._check_contains,
        ]:
            try:
                checker(copy, original=other)
            except NoMatch:
                return False

        return True

    def __str__(self):
        # This is some pretty gross code, but it makes test output so much
        # more readable
        parts = ["any"]

        parts.extend(self._describe_type())
        parts.extend(self._describe_size())
        parts.extend(self._describe_contains())
        parts.extend(self._describe_item_matcher())

        return f'* {" ".join(parts)} *'


class AnyDict(AnyCollection):
    """A matcher representing any dict."""

    _exact_type = dict


class AnySet(AnyCollection):
    """A matcher representing any set."""

    _exact_type = set


class AnyList(AnyCollection):
    """A matcher representing any list."""

    _exact_type = list


class AnyMapping(AnyCollection):
    """A matcher representing any mapping."""

    def __eq__(self, other):
        if not hasattr(other, "items"):
            return False

        return super().__eq__(other)



"""Matchers formed of combinations of other things."""
from h_matchers.matcher.core import Matcher

# pylint: disable=too-few-public-methods


class AnyOf(Matcher):
    """Match any one of a series of options."""

    def __init__(self, options):
        options = list(options)  # Coerce generators into concrete list

        super().__init__(
            f"* any of {options} *", lambda other: other in options,
        )


class AllOf(Matcher):
    """Match only when all of a series of options match."""

    def __init__(self, options):
        options = list(options)  # Coerce generators into concrete list

        super().__init__(
            f"* all of {options} *",
            lambda other: all(other == option for option in options),
        )


class NamedMatcher(Matcher):
    """Wrap a matcher with a custom description for nice stringification."""

    def __init__(self, description, matcher):
        super().__init__(description, matcher.__eq__)

    def __repr__(self):
        return self._description



"""A matcher that matches anything."""
from h_matchers.matcher.core import Matcher

# pylint: disable=too-few-public-methods


class AnyThing(Matcher):
    """Matches anything."""

    def __init__(self):
        super().__init__("* anything *", lambda _: True)



"""A matcher that matches various request objects in HTTP type libraries."""
from copy import deepcopy

from h_matchers.decorator import fluent_entrypoint
from h_matchers.matcher.collection import AnyMapping
from h_matchers.matcher.core import Matcher
from h_matchers.matcher.strings import AnyString
from h_matchers.matcher.web.url import AnyURL

# pylint: disable=too-few-public-methods


class _LibraryNotAvailable:
    """Substitute for missing libraries."""


try:  # pylint: disable=too-many-try-statements
    # We don't require our clients to install every request library we support
    # so we will enable it if it's around, and carry on without if it's not
    from requests import PreparedRequest as RequestsPreparedRequest
    from requests import Request as RequestsRequest

except ImportError:  # pragma: no cover
    RequestsRequest, RequestsPreparedRequest = (
        _LibraryNotAvailable,
        _LibraryNotAvailable,
    )

try:  # pylint: disable=too-many-try-statements
    from pyramid.request import Request as PyramidRequest
    from pyramid.testing import DummyRequest as PyramidDummyRequest

except ImportError:  # pragma: no cover
    PyramidDummyRequest, PyramidRequest = _LibraryNotAvailable, _LibraryNotAvailable


# pylint: disable=function-redefined


class AnyRequest(Matcher):  # pragma: no cover
    """Matching object for request type objects.

    Currently supported request objects:

     * `requests.Request`
     * `requests.PreparedRequest`
     * `pyramid.request.Request`
     * `pyramid.testing.Request`
    """

    # Use a set here as an optimisation for when many of these items are
    # instances of `_LibraryNotAvailable`
    SUPPORTED_TYPES = tuple(
        {RequestsRequest, RequestsPreparedRequest, PyramidRequest, PyramidDummyRequest}
    )

    # Enable failing on comparison instead of returning False. This can be very
    # useful for debugging as we can fail fast and return a message about why
    # we can't match. We might want to think about making this a more general
    # feature. It would be very handy for the more complex matchers
    _assert_on_comparison = False

    method = None
    url = None
    headers = None

    def __init__(self, method=..., url=..., headers=None):
        self.with_method(method)
        self.with_url(url)
        self.with_headers(headers)

        super().__init__("*dummy*", self._matches_request)

    @classmethod
    def containing_headers(cls, headers):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @classmethod
    def with_headers(cls, header=...):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @classmethod
    def with_method(cls, method=...):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @classmethod
    def with_url(cls, url=...):
        """Confuse pylint so it doesn't complain about fluent-endpoints."""

    @fluent_entrypoint
    def containing_headers(self, headers):
        """Specify the request must have at least the headers specified.

        :param headers: A mappable of headers to match
        """
        self.headers = AnyMapping.containing(headers)

    @fluent_entrypoint
    def with_headers(self, headers=...):
        """Specify the request must have headers.

        This method will remove 'Host' from the set of headers as some
        libraries add it before sending, and it's mostly noise.

        :param headers: A mappable of headers or matcher to match exactly
        """
        if headers is ...:
            self.headers = AnyMapping.of_size(at_least=1)

        elif isinstance(headers, Matcher):
            self.headers = headers

        elif headers is not None:
            self.headers = AnyMapping.containing(headers).only()

    @fluent_entrypoint
    def with_method(self, method=...):
        """Specify the request must have a method.

        :param method: A string or matcher for the method
        """
        if method is ...:
            self.method = AnyString()

        elif isinstance(method, str):
            self.method = method.upper()

        elif method is not None:
            self.method = method

    @fluent_entrypoint
    def with_url(self, url=...):
        """Specify the request must have a URL.

        :param url: A string or matcher for the URL
        """
        if url is ...:
            self.url = AnyURL()

        elif isinstance(url, str):
            self.url = AnyURL.matching(url)

        elif url is not None:
            self.url = url

    def assert_equal_to(self, other):
        """Assert that the request object is equal to another object.

        :raise AssertionError: If no match is found with details of why
        """
        if not isinstance(other, self.SUPPORTED_TYPES):
            raise AssertionError(
                f"Unknown request type '{type(other)}'. For a request type to be "
                "compared it must be supported and loaded."
            )

        if self.method is not None and self.method != other.method.upper():
            raise AssertionError(f"Method '{other.method}' != '{self.method}'")

        if self.url is not None and self.url != other.url:
            raise AssertionError(f"URL '{other.url}' != '{self.url}'")

        if self.headers is not None:
            other_headers = self._comparison_headers(other)
            if self.headers != other_headers:
                raise AssertionError(f"Headers {other_headers} != {self.headers}")

    @classmethod
    def _comparison_headers(cls, other):
        if isinstance(other, PyramidRequest):
            headers = deepcopy(other.headers)
            headers.pop("Host")

            return headers

        return other.headers

    def _matches_request(self, other):
        try:
            self.assert_equal_to(other)
        except AssertionError:
            if self._assert_on_comparison:
                raise
            return False

        return True

    def __str__(self):
        details = ""
        if self.method:
            details += f" method:{self.method}"

        if self.url:
            details += f" url:{self.url}"

        if self.headers is not None:
            details += f" headers={self.headers}"

        return f"<{self.__class__.__name__}{details}>"
